/***************************************************************************
                          chublistmanager.cpp  -  description
                             -------------------
    begin                : Don Mai 16 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@users.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cclient.h"
#include "cconfig.h"
#include "core/cmanager.h"
#include "dcobject.h"
#include "core/cbytearray.h"
#include "chttp.h"
#include "core/cbz.h"
#include "cutils.h"

#include "core/ciconv.h"

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include "chublistmanager.h"

/** */
CHubListManager::CHubListManager()
{
	m_pCallback = new CCallback<CHubListManager>( this, &CHubListManager::Callback );
	CManager::Instance()->Add( m_pCallback );

	if ( CConfig::Instance()->GetReloadHubListTime() != 0 )
	{
		m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
	}
	else
	{
		m_nReloadHubListTimeout = 0;
	}

	// get hublist stuff
	m_pHttp           = 0;
	m_pHubListUrlList = 0;
	m_pHubListUrl     = 0;
	m_pHubListData    = 0;

	m_bGetHubListDone = FALSE;
	
	m_pXml = 0;
	m_pXmlHubs = new CList<DCConfigHubItem>;
}

/** */
CHubListManager::~CHubListManager()
{
	m_Thread.Stop();

	SetInstance(0);

	CManager::Instance()->Remove( m_pCallback );

	if (m_pCallback)
	{
		delete m_pCallback;
		m_pCallback = 0;
	}
	
	if (m_pXml != 0)
	{
		delete m_pXml;
		m_pXml = 0;
	}
	
	if (m_pXmlHubs != 0)
	{
		delete m_pXmlHubs;
		m_pXmlHubs = 0;
	}
}

/** thread callbackfunction */
int CHubListManager::Callback( CObject *, CObject * )
{
	m_Thread.Lock();

	// check reload hublist timeout
	if ( CConfig::Instance() )
	{
		if ( CConfig::Instance()->GetReloadHubListTime() != 0 )
		{
			if ( m_nReloadHubListTimeout != 0 )
			{
				if ( time(0) >= m_nReloadHubListTimeout )
				{
					GetPublicHubList();

					m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
				}
			}
			else // change from config
			{
				m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
			}
		}
		else
		{
			m_nReloadHubListTimeout = 0;
		}
	}

	// cleanup all objects
	if ( m_bGetHubListDone == TRUE )
	{
		if ( m_pHttp )
		{
			delete m_pHttp;
			m_pHttp = 0;
		}

		if ( m_pHubListUrlList )
		{
			delete m_pHubListUrlList;
			m_pHubListUrlList = 0;
		}

		if ( m_pHubListData )
		{
			delete m_pHubListData;
			m_pHubListData = 0;
		}

		if ( m_pXmlHubs )
		{
			m_pXmlHubs->Clear();
		}

		m_pHubListUrl = 0;

		DCMessageGetHubList * msg = new DCMessageGetHubList();

		msg->m_bRun = FALSE;

		if ( DC_CallBack(msg) == -1 )
		{
			delete msg;
		}

		m_bGetHubListDone = FALSE;
	}

	m_Thread.UnLock();

	return 0;
}

/** http callback function */
int CHubListManager::HttpCallBack( CObject *, CObject * object )
{
	CByteArray in;

	switch(((CDCMessage*)object)->m_eType)
	{
		case DC_MESSAGE_CONNECTION_STATE:
		{
			CMessageConnectionState *msg = (CMessageConnectionState*)object;

			if ( msg->m_eState == estDISCONNECTED )
			{
				if ( (m_pHttp->GetHttpError() == 200) && (m_pHttp->GetData(&in) == TRUE) )
				{
					HandleHubListData( m_pHttp->GetUrl(), &in );
				}

				// redirect
				if ( m_pHttp->GetHttpError() == 302 )
				{
					m_pHttp->GetUrl(m_pHttp->GetLocation());
				}
				else if ( NextHubListUrl() == FALSE )
				{
					// parse public hub list in own thread
					CCallback<CHubListManager> * c = new CCallback<CHubListManager>( this, &CHubListManager::ParsePublicHubList );
					m_Thread.SetThreadCallBackFunction(c);
					m_Thread.Start();
				}
			}

			break;
		}

		case DC_MESSAGE_TRANSFER:
		{
			if ( DC_CallBack(object) != -1 )
			{
				object = 0;
			}

			break;
		}

		default:
		{
			break;
		}
	}

	if ( object )
	{
		delete object;
	}

	return 0;
}

/** */
void CHubListManager::HandleHubListData( const CString & url, CByteArray * in )
{
	CByteArray * fixed = 0;
	CByteArray out;
	CBZ bz2;
	
	if ( url.Right(4).ToLower() == ".bz2" )
	{
		if ( bz2.Decompress( in, &out ) == TRUE )
		{
			if ( url.Right(8).ToLower() == ".xml.bz2" )
			{
				if ( m_pXml == 0 )
				{
					m_pXml = new CXml();
				}
				
				fixed = RemoveInvalidXmlChars(&out);
				
				if ( m_pXml->ParseMemory((const char*)fixed->Data(),fixed->Size()) == TRUE )
				{
					ParseXmlPublicHubList();
				}
				else
				{
					printf("Failed to parse XML hublist.\n");
				}
				
				delete m_pXml;
				m_pXml = 0;
				
				delete fixed;
			}
			else
			{
				m_pHubListData->Append(out.Data(),out.Size());
				m_pHubListData->Append("\xD\xA",2);
			}
		}
		else
		{
			printf("bz2 decompress failed\n");
		}
	}
	else
	{
		if ( url.Right(4).ToLower() == ".xml" )
		{
			if ( m_pXml == 0 )
			{
				m_pXml = new CXml();
			}
			
			fixed = RemoveInvalidXmlChars(in);
			
			if ( m_pXml->ParseMemory((const char*)fixed->Data(),fixed->Size()) == TRUE )
			{
				ParseXmlPublicHubList();
			}
			else
			{
				printf("Failed to parse XML hublist.\n");
			}
			
			delete m_pXml;
			m_pXml = 0;
			
			delete fixed;
		}
		else
		{
			m_pHubListData->Append(in->Data(),in->Size());
			m_pHubListData->Append("\xD\xA",2);
		}
	}
}

/** */
bool CHubListManager::GetPublicHubList()
{
	bool res = FALSE;

	if ( m_pHttp != 0 )
	{
		// get hublist allready running
		return res;
	}

	m_pHubListUrlList = new CList<DCConfigHubListUrl>();

	// get all hublist urls
	CConfig::Instance()->GetHubListUrlList(m_pHubListUrlList);

	// check hublist url count
	if ( m_pHubListUrlList->Count() == 0 )
	{
		delete m_pHubListUrlList;
		m_pHubListUrlList = 0;
		return res;
	}

	// init first url
	m_pHubListUrl = 0;

	m_pHubListData = new CByteArray();
	m_pHttp = new CHttp();
	m_pHttp->SetCallBackFunction( new CCallback<CHubListManager>( this, &CHubListManager::HttpCallBack ) );

	res = NextHubListUrl();

	if ( !res )
	{
		m_bGetHubListDone = TRUE;
	}
	else
	{
		DCMessageGetHubList * msg = new DCMessageGetHubList();

		msg->m_bRun = TRUE;

		if ( DC_CallBack(msg) == -1 )
		{
			delete msg;
		}
	}

	return res;
}

/** */
int CHubListManager::ParsePublicHubList( CObject * , CObject * )
{
	CString line;
	CString s;
	CString s1,s2,s3,s4;
	long i=0,i1=0;

	DCConfigHubItem * hubitem = 0;
	
	if ( (m_pHubListData->Size() == 0) && (m_pXmlHubs->Count() == 0) )
	{
		m_Thread.Stop();
		m_Thread.SetThreadCallBackFunction(0);
		m_bGetHubListDone = TRUE;

		return 0;
	}
	
	while ( (hubitem = m_pXmlHubs->Next(hubitem)) != 0 )
	{
		// name, host, description, usercount, country, shared, minshare, extra
		CConfig::Instance()->AddPublicHub(
			hubitem->m_sName,
			hubitem->m_sHost,
			hubitem->m_sDescription,
			hubitem->m_sUserCount,
			hubitem->m_sCountry,
			hubitem->m_nShared,
			hubitem->m_nMinShare,
			hubitem->m_sExtra
		);
	}
	
	if ( m_pHubListData->Size() > 0 )
	{
	
	s.Set((const char*)m_pHubListData->Data(),m_pHubListData->Size());

	CIconv * pIconv = new CIconv( CConfig::Instance()->GetRemoteEncoding(), CConfig::Instance()->GetLocalEncoding() );

	while( (i = s.Find(0x0d,i)) != -1 )
	{
		line = s.Mid(i1,i-i1);

		if ( !line.IsEmpty() )
		{
			s1 = line.Section( '|', 0, 0 );
			s2 = line.Section( '|', 1, 1 );
			s3 = line.Section( '|', 2, 2 );
			s4 = line.Section( '|', 3, 3 );

			// replace all spaces
			s2 = s2.Replace(" ","");

			// remove line ends
			s1 = s1.Replace("\n", "");

			// name, host, description, usercount
			CConfig::Instance()->AddPublicHub( pIconv->encode(s1), pIconv->encode(s2), pIconv->encode(s3), s4 );
		}

		i1 = i+2;
		i += 2;
	}

	delete pIconv;
	
	}

	// store the list
	if ( CConfig::Instance()->GetHubListStoreLocal() == TRUE )
	{
		CConfig::Instance()->SaveDCHub();
	}

	m_Thread.Stop();
	m_Thread.SetThreadCallBackFunction(0);
	m_bGetHubListDone = TRUE;

	return 0;
}

/** */
bool CHubListManager::NextHubListUrl()
{
	bool res = FALSE;

	while( (m_pHubListUrl=m_pHubListUrlList->Next(m_pHubListUrl)) != 0 )
	{
		if ( m_pHubListUrl->bEnabled == TRUE )
		{
			if ( m_pHubListUrl->sUrl.Left(7) == "file://" )
			{
				CByteArray * data = new CByteArray();
				if ( data->LoadFromFile(m_pHubListUrl->sUrl.Mid(7)) == TRUE )
				{
					HandleHubListData( m_pHubListUrl->sUrl, data );
				}
				delete data;
				
				if ( NextHubListUrl() == FALSE )
				{
					// parse public hub list in own thread
					CCallback<CHubListManager> * c = new CCallback<CHubListManager>( this, &CHubListManager::ParsePublicHubList );
					m_Thread.SetThreadCallBackFunction(c);
					m_Thread.Start();
				}
				
				res = TRUE;
				break;
			}
			else if ( m_pHubListUrl->sUrl.NotEmpty() )
			{
				m_pHttp->GetUrl(m_pHubListUrl->sUrl);

				res = TRUE;
				break;
			}
		}
	}

	return res;
}

/** */
int CHubListManager::ParseXmlPublicHubList()
{
	xmlNodePtr node, node2;
	int count = 0;
	
	printf("Parse XML hub list...\n");
	
	for ( node=m_pXml->doc()->children; node != 0; node=node->next )
	{
		if ( m_pXml->name(node) == "Hublist" )
		{
			CList<CXmlColumn> * cols = FindAndParseXmlColumns( node->children );
			
			if ( !cols )
			{
				/*
				 * The column headings are only needed so that the Extra
				 * field is filled in.
				 */
				printf("ParseXmlPublicHubList: no column headings found, trying with defaults\n");
				
				cols = new CList<CXmlColumn>;
				
				CXmlColumn * xmlcol = new CXmlColumn();
				xmlcol->m_sName = "Name";
				xmlcol->m_sType = "string";
				cols->Add(xmlcol);
				
				xmlcol = new CXmlColumn();
				xmlcol->m_sName = "Address";
				xmlcol->m_sType = "string";
				cols->Add(xmlcol);
				
				xmlcol = new CXmlColumn();
				xmlcol->m_sName = "Description";
				xmlcol->m_sType = "string";
				cols->Add(xmlcol);

				xmlcol = new CXmlColumn();
				xmlcol->m_sName = "Port";
				xmlcol->m_sType = "int";
				cols->Add(xmlcol);
				
				xmlcol = new CXmlColumn();
				xmlcol->m_sName = "Users";
				xmlcol->m_sType = "int";
				cols->Add(xmlcol);
			}
			
			for ( node2 = node->children; node2 != 0; node2=node2->next )
			{
				if ( m_pXml->name(node2) == "Hubs" )
				{
					count += ParseXmlHubs( node2->children, cols );
				}
			}
			
			cols->Clear();
			delete cols;
		}
	}
	
	printf("XML hublist: %d hubs\n", count);
	
	return count;
}

/** */
int CHubListManager::ParseXmlHubs( xmlNodePtr node, CList<CXmlColumn> * cols )
{
	xmlNodePtr n1;
	int count = 0;
	
	for ( n1=node; n1 != 0; n1=n1->next )
	{
		if ( m_pXml->name(n1) == "Hub" )
		{
			ParseXmlHub(n1, cols);
			count++;
		}
	}
	
	return count;
}

/** */
CList<CXmlColumn> * CHubListManager::FindAndParseXmlColumns( xmlNodePtr node )
{
	xmlNodePtr n1,n2,n3;
	
	CList<CXmlColumn> * cols = 0;
	
	for ( n1=node; n1 != 0; n1=n1->next )
	{
		if ( m_pXml->name(n1) == "Columns" )
		{
			cols = new CList<CXmlColumn>;
			for ( n2 = n1->children; n2 != 0; n2 = n2->next )
			{
				if ( m_pXml->name(n2) == "Column" )
				{
					CXmlColumn * col = new CXmlColumn();
					col->m_sName = m_pXml->prop(n2,"Name");
					col->m_sType = m_pXml->prop(n2,"Type");
					cols->Add(col);
				}
			}
			break;
		}
		
		for ( n3 = n1->children; n3 != 0; n3 = n3->next )
		{
			if ( m_pXml->name(n3) == "Columns" )
			{
				cols = new CList<CXmlColumn>;
				for ( n2 = n3->children; n2 != 0; n2 = n2->next )
				{
					if ( m_pXml->name(n2) == "Column" )
					{
						CXmlColumn * col = new CXmlColumn();
						col->m_sName = m_pXml->prop(n2,"Name");
						col->m_sType = m_pXml->prop(n2,"Type");
						cols->Add(col);
					}
				}
				break;
			}
		}
		
		if ( cols )
		{
			break;
		}
	}
	
	return cols;
}

/** */
void CHubListManager::ParseXmlHub( xmlNodePtr node, CList<CXmlColumn> * cols )
{
	CXmlColumn * col = 0;
	DCConfigHubItem * hubitem = new DCConfigHubItem();
	CString port;
	
	while ( (col = cols->Next(col)) != 0 )
	{
		col->m_sValue = m_pXml->prop(node, col->m_sName);
		
		if (col->m_sName.ToLower() == "name")
		{
			hubitem->m_sName = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "address")
		{
			hubitem->m_sHost = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "description")
		{
			hubitem->m_sDescription = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "users")
		{
			hubitem->m_sUserCount = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "port")
		{
			port = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "country")
		{
			hubitem->m_sCountry = col->m_sValue;
		}
		else if (col->m_sName.ToLower() == "minshare")
		{
			hubitem->m_nMinShare = col->m_sValue.asULL();
		}
		else if (col->m_sName.ToLower() == "shared")
		{
			hubitem->m_nShared = col->m_sValue.asULL();
		}
		else
		{
			hubitem->m_sExtra += col->m_sName;
			hubitem->m_sExtra += '=';
			hubitem->m_sExtra += col->m_sValue;
			hubitem->m_sExtra += ' ';
		}
		
		//printf("Name=%s Type=%s Value=%s\n", col->m_sName.Data(), col->m_sType.Data(), col->m_sValue.Data());
	}
	
	if ( (hubitem->m_sHost.Find(':') < 0) && (port.NotEmpty()) )
	{
		hubitem->m_sHost += ':';
		hubitem->m_sHost += port;
	}
	
	if ( hubitem->m_sName.IsEmpty() || hubitem->m_sHost.IsEmpty() )
	{
		delete hubitem;
	}
	else
	{
		m_pXmlHubs->Add(hubitem);
	}
}

/** */
CByteArray * CHubListManager::RemoveInvalidXmlChars( CByteArray * input )
{
	/*
	 * OK although we need to check the characters one by one we do not
	 * want to copy them one by one because that seems excessively slow.
	 */
	CByteArray * output = new CByteArray(0);
	
	long goodstart = 0;
	unsigned char c;
	long bad = 0;
	
	/*
	 * Although all the XML hublists I checked were not null terminated, and all had
	 * different last characters, the last character is probably something special
	 * and in any case is highly unlikely to upset libxml2.
	 */
	long len = input->Size() - 1;
	
	/*
	 * I am assuming all & ' " < > characters are properly escaped.
	 * Alternatively the data could be run through CXml::UnEscapeSpecials and then
	 * through CXml::EscapeSpecials, but again, that sounds too slow.
	 * The ranges of invalid characters is taken from CXml::EscapeSpecials.
	 */
	for ( long i = 0; i < len; i++ )
	{
		c = input->Data()[i];
		
		// c >= 0x00 always true ...
		
		if ( ( c <= 0x08 ) ||
		     ( c == 0x0B ) ||
		     ( c == 0x0C ) ||
		     ( c >= 0x0E && c <= 0x1F )
		   )
		{
			bad++;
			/* bad character encountered
			 * first copy all not yet copied valid data to the output
			 * but skip the copy in case there is more than one bad character
			 * in a row
			 */ 
			long size = i-goodstart;
			//printf("Bad char at %ld, input->Data()=%ld, input->Data()+goodstart=%ld, size=%ld\n",i,input->Data(),input->Data()+goodstart,size);
			if ( size > 0 )
			{
				output->Append( input->Data()+goodstart, size );
			}
			
			/* for whatever reason, this appears to be incorrect even though it is
			 * what CXml::EscapeSpecials does, e.g. &#00 generates
			 *
			 * xmlParseCharRef: invalid xmlChar value 0
			 *
			 * so just skip the damn invalid characters
			 *
			 * CString escape = "&#0";
			 * escape += CString().setNum(c);
			 * escape += ';';
			 *
			 * output->Append( escape.Data(), escape.Length() );
			 */
			
			// update goodstart
			goodstart = i + 1;
		}
	}
	
	/* Copy remaining data, which ranges from all the data if no invalid
	 * characters were found to just the last byte if the last but one character
	 * was invalid.
	 */
	output->Append( input->Data()+goodstart, input->Size()-goodstart );
	
	if ( bad > 0 )
	{
		printf("CHubListManager: %ld invalid characters were removed from the XML hublist\n",bad);
	}
	
	/* For testing. */
	//output->SaveToFile("/var/tmp/dclib-fixedup.xml");
	
	return output;
}
